// Code generated by "stringer -type=EventType"; DO NOT EDIT.

package tui

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[Rune-0]
	_ = x[CtrlA-1]
	_ = x[CtrlB-2]
	_ = x[CtrlC-3]
	_ = x[CtrlD-4]
	_ = x[CtrlE-5]
	_ = x[CtrlF-6]
	_ = x[CtrlG-7]
	_ = x[CtrlH-8]
	_ = x[Tab-9]
	_ = x[CtrlJ-10]
	_ = x[CtrlK-11]
	_ = x[CtrlL-12]
	_ = x[Enter-13]
	_ = x[CtrlN-14]
	_ = x[CtrlO-15]
	_ = x[CtrlP-16]
	_ = x[CtrlQ-17]
	_ = x[CtrlR-18]
	_ = x[CtrlS-19]
	_ = x[CtrlT-20]
	_ = x[CtrlU-21]
	_ = x[CtrlV-22]
	_ = x[CtrlW-23]
	_ = x[CtrlX-24]
	_ = x[CtrlY-25]
	_ = x[CtrlZ-26]
	_ = x[Esc-27]
	_ = x[CtrlSpace-28]
	_ = x[CtrlDelete-29]
	_ = x[CtrlBackSlash-30]
	_ = x[CtrlRightBracket-31]
	_ = x[CtrlCaret-32]
	_ = x[CtrlSlash-33]
	_ = x[ShiftTab-34]
	_ = x[Backspace-35]
	_ = x[Delete-36]
	_ = x[PageUp-37]
	_ = x[PageDown-38]
	_ = x[Up-39]
	_ = x[Down-40]
	_ = x[Left-41]
	_ = x[Right-42]
	_ = x[Home-43]
	_ = x[End-44]
	_ = x[Insert-45]
	_ = x[ShiftUp-46]
	_ = x[ShiftDown-47]
	_ = x[ShiftLeft-48]
	_ = x[ShiftRight-49]
	_ = x[ShiftDelete-50]
	_ = x[F1-51]
	_ = x[F2-52]
	_ = x[F3-53]
	_ = x[F4-54]
	_ = x[F5-55]
	_ = x[F6-56]
	_ = x[F7-57]
	_ = x[F8-58]
	_ = x[F9-59]
	_ = x[F10-60]
	_ = x[F11-61]
	_ = x[F12-62]
	_ = x[AltBackspace-63]
	_ = x[AltUp-64]
	_ = x[AltDown-65]
	_ = x[AltLeft-66]
	_ = x[AltRight-67]
	_ = x[AltShiftUp-68]
	_ = x[AltShiftDown-69]
	_ = x[AltShiftLeft-70]
	_ = x[AltShiftRight-71]
	_ = x[Alt-72]
	_ = x[CtrlAlt-73]
	_ = x[Invalid-74]
	_ = x[Fatal-75]
	_ = x[Mouse-76]
	_ = x[DoubleClick-77]
	_ = x[LeftClick-78]
	_ = x[RightClick-79]
	_ = x[SLeftClick-80]
	_ = x[SRightClick-81]
	_ = x[ScrollUp-82]
	_ = x[ScrollDown-83]
	_ = x[SScrollUp-84]
	_ = x[SScrollDown-85]
	_ = x[PreviewScrollUp-86]
	_ = x[PreviewScrollDown-87]
	_ = x[Resize-88]
	_ = x[Change-89]
	_ = x[BackwardEOF-90]
	_ = x[Start-91]
	_ = x[Load-92]
	_ = x[Focus-93]
	_ = x[One-94]
	_ = x[Zero-95]
	_ = x[Result-96]
	_ = x[Jump-97]
	_ = x[JumpCancel-98]
	_ = x[ClickHeader-99]
}

const _EventType_name = "RuneCtrlACtrlBCtrlCCtrlDCtrlECtrlFCtrlGCtrlHTabCtrlJCtrlKCtrlLEnterCtrlNCtrlOCtrlPCtrlQCtrlRCtrlSCtrlTCtrlUCtrlVCtrlWCtrlXCtrlYCtrlZEscCtrlSpaceCtrlDeleteCtrlBackSlashCtrlRightBracketCtrlCaretCtrlSlashShiftTabBackspaceDeletePageUpPageDownUpDownLeftRightHomeEndInsertShiftUpShiftDownShiftLeftShiftRightShiftDeleteF1F2F3F4F5F6F7F8F9F10F11F12AltBackspaceAltUpAltDownAltLeftAltRightAltShiftUpAltShiftDownAltShiftLeftAltShiftRightAltCtrlAltInvalidFatalMouseDoubleClickLeftClickRightClickSLeftClickSRightClickScrollUpScrollDownSScrollUpSScrollDownPreviewScrollUpPreviewScrollDownResizeChangeBackwardEOFStartLoadFocusOneZeroResultJumpJumpCancelClickHeader"

var _EventType_index = [...]uint16{0, 4, 9, 14, 19, 24, 29, 34, 39, 44, 47, 52, 57, 62, 67, 72, 77, 82, 87, 92, 97, 102, 107, 112, 117, 122, 127, 132, 135, 144, 154, 167, 183, 192, 201, 209, 218, 224, 230, 238, 240, 244, 248, 253, 257, 260, 266, 273, 282, 291, 301, 312, 314, 316, 318, 320, 322, 324, 326, 328, 330, 333, 336, 339, 351, 356, 363, 370, 378, 388, 400, 412, 425, 428, 435, 442, 447, 452, 463, 472, 482, 492, 503, 511, 521, 530, 541, 556, 573, 579, 585, 596, 601, 605, 610, 613, 617, 623, 627, 637, 648}

func (i EventType) String() string {
	if i < 0 || i >= EventType(len(_EventType_index)-1) {
		return "EventType(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _EventType_name[_EventType_index[i]:_EventType_index[i+1]]
}
