// Code generated by "stringer -type=actionType"; DO NOT EDIT.

package fzf

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[actIgnore-0]
	_ = x[actStart-1]
	_ = x[actClick-2]
	_ = x[actInvalid-3]
	_ = x[actChar-4]
	_ = x[actMouse-5]
	_ = x[actBeginningOfLine-6]
	_ = x[actAbort-7]
	_ = x[actAccept-8]
	_ = x[actAcceptNonEmpty-9]
	_ = x[actAcceptOrPrintQuery-10]
	_ = x[actBackwardChar-11]
	_ = x[actBackwardDeleteChar-12]
	_ = x[actBackwardDeleteCharEof-13]
	_ = x[actBackwardWord-14]
	_ = x[actCancel-15]
	_ = x[actChangeBorderLabel-16]
	_ = x[actChangeListLabel-17]
	_ = x[actChangeInputLabel-18]
	_ = x[actChangeHeader-19]
	_ = x[actChangeHeaderLabel-20]
	_ = x[actChangeMulti-21]
	_ = x[actChangePreviewLabel-22]
	_ = x[actChangePrompt-23]
	_ = x[actChangeQuery-24]
	_ = x[actChangeNth-25]
	_ = x[actClearScreen-26]
	_ = x[actClearQuery-27]
	_ = x[actClearSelection-28]
	_ = x[actClose-29]
	_ = x[actDeleteChar-30]
	_ = x[actDeleteCharEof-31]
	_ = x[actEndOfLine-32]
	_ = x[actFatal-33]
	_ = x[actForwardChar-34]
	_ = x[actForwardWord-35]
	_ = x[actKillLine-36]
	_ = x[actKillWord-37]
	_ = x[actUnixLineDiscard-38]
	_ = x[actUnixWordRubout-39]
	_ = x[actYank-40]
	_ = x[actBackwardKillWord-41]
	_ = x[actSelectAll-42]
	_ = x[actDeselectAll-43]
	_ = x[actToggle-44]
	_ = x[actToggleSearch-45]
	_ = x[actToggleAll-46]
	_ = x[actToggleDown-47]
	_ = x[actToggleUp-48]
	_ = x[actToggleIn-49]
	_ = x[actToggleOut-50]
	_ = x[actToggleTrack-51]
	_ = x[actToggleTrackCurrent-52]
	_ = x[actToggleHeader-53]
	_ = x[actToggleWrap-54]
	_ = x[actToggleMultiLine-55]
	_ = x[actToggleHscroll-56]
	_ = x[actTrackCurrent-57]
	_ = x[actToggleInput-58]
	_ = x[actHideInput-59]
	_ = x[actShowInput-60]
	_ = x[actUntrackCurrent-61]
	_ = x[actDown-62]
	_ = x[actUp-63]
	_ = x[actPageUp-64]
	_ = x[actPageDown-65]
	_ = x[actPosition-66]
	_ = x[actHalfPageUp-67]
	_ = x[actHalfPageDown-68]
	_ = x[actOffsetUp-69]
	_ = x[actOffsetDown-70]
	_ = x[actOffsetMiddle-71]
	_ = x[actJump-72]
	_ = x[actJumpAccept-73]
	_ = x[actPrintQuery-74]
	_ = x[actRefreshPreview-75]
	_ = x[actReplaceQuery-76]
	_ = x[actToggleSort-77]
	_ = x[actShowPreview-78]
	_ = x[actHidePreview-79]
	_ = x[actTogglePreview-80]
	_ = x[actTogglePreviewWrap-81]
	_ = x[actTransform-82]
	_ = x[actTransformBorderLabel-83]
	_ = x[actTransformListLabel-84]
	_ = x[actTransformInputLabel-85]
	_ = x[actTransformHeader-86]
	_ = x[actTransformHeaderLabel-87]
	_ = x[actTransformNth-88]
	_ = x[actTransformPreviewLabel-89]
	_ = x[actTransformPrompt-90]
	_ = x[actTransformQuery-91]
	_ = x[actTransformSearch-92]
	_ = x[actSearch-93]
	_ = x[actPreview-94]
	_ = x[actChangePreview-95]
	_ = x[actChangePreviewWindow-96]
	_ = x[actPreviewTop-97]
	_ = x[actPreviewBottom-98]
	_ = x[actPreviewUp-99]
	_ = x[actPreviewDown-100]
	_ = x[actPreviewPageUp-101]
	_ = x[actPreviewPageDown-102]
	_ = x[actPreviewHalfPageUp-103]
	_ = x[actPreviewHalfPageDown-104]
	_ = x[actPrevHistory-105]
	_ = x[actPrevSelected-106]
	_ = x[actPrint-107]
	_ = x[actPut-108]
	_ = x[actNextHistory-109]
	_ = x[actNextSelected-110]
	_ = x[actExecute-111]
	_ = x[actExecuteSilent-112]
	_ = x[actExecuteMulti-113]
	_ = x[actSigStop-114]
	_ = x[actFirst-115]
	_ = x[actLast-116]
	_ = x[actReload-117]
	_ = x[actReloadSync-118]
	_ = x[actDisableSearch-119]
	_ = x[actEnableSearch-120]
	_ = x[actSelect-121]
	_ = x[actDeselect-122]
	_ = x[actUnbind-123]
	_ = x[actRebind-124]
	_ = x[actToggleBind-125]
	_ = x[actBecome-126]
	_ = x[actShowHeader-127]
	_ = x[actHideHeader-128]
	_ = x[actBell-129]
	_ = x[actExclude-130]
	_ = x[actExcludeMulti-131]
}

const _actionType_name = "actIgnoreactStartactClickactInvalidactCharactMouseactBeginningOfLineactAbortactAcceptactAcceptNonEmptyactAcceptOrPrintQueryactBackwardCharactBackwardDeleteCharactBackwardDeleteCharEofactBackwardWordactCancelactChangeBorderLabelactChangeListLabelactChangeInputLabelactChangeHeaderactChangeHeaderLabelactChangeMultiactChangePreviewLabelactChangePromptactChangeQueryactChangeNthactClearScreenactClearQueryactClearSelectionactCloseactDeleteCharactDeleteCharEofactEndOfLineactFatalactForwardCharactForwardWordactKillLineactKillWordactUnixLineDiscardactUnixWordRuboutactYankactBackwardKillWordactSelectAllactDeselectAllactToggleactToggleSearchactToggleAllactToggleDownactToggleUpactToggleInactToggleOutactToggleTrackactToggleTrackCurrentactToggleHeaderactToggleWrapactToggleMultiLineactToggleHscrollactTrackCurrentactToggleInputactHideInputactShowInputactUntrackCurrentactDownactUpactPageUpactPageDownactPositionactHalfPageUpactHalfPageDownactOffsetUpactOffsetDownactOffsetMiddleactJumpactJumpAcceptactPrintQueryactRefreshPreviewactReplaceQueryactToggleSortactShowPreviewactHidePreviewactTogglePreviewactTogglePreviewWrapactTransformactTransformBorderLabelactTransformListLabelactTransformInputLabelactTransformHeaderactTransformHeaderLabelactTransformNthactTransformPreviewLabelactTransformPromptactTransformQueryactTransformSearchactSearchactPreviewactChangePreviewactChangePreviewWindowactPreviewTopactPreviewBottomactPreviewUpactPreviewDownactPreviewPageUpactPreviewPageDownactPreviewHalfPageUpactPreviewHalfPageDownactPrevHistoryactPrevSelectedactPrintactPutactNextHistoryactNextSelectedactExecuteactExecuteSilentactExecuteMultiactSigStopactFirstactLastactReloadactReloadSyncactDisableSearchactEnableSearchactSelectactDeselectactUnbindactRebindactToggleBindactBecomeactShowHeaderactHideHeaderactBellactExcludeactExcludeMulti"

var _actionType_index = [...]uint16{0, 9, 17, 25, 35, 42, 50, 68, 76, 85, 102, 123, 138, 159, 183, 198, 207, 227, 245, 264, 279, 299, 313, 334, 349, 363, 375, 389, 402, 419, 427, 440, 456, 468, 476, 490, 504, 515, 526, 544, 561, 568, 587, 599, 613, 622, 637, 649, 662, 673, 684, 696, 710, 731, 746, 759, 777, 793, 808, 822, 834, 846, 863, 870, 875, 884, 895, 906, 919, 934, 945, 958, 973, 980, 993, 1006, 1023, 1038, 1051, 1065, 1079, 1095, 1115, 1127, 1150, 1171, 1193, 1211, 1234, 1249, 1273, 1291, 1308, 1326, 1335, 1345, 1361, 1383, 1396, 1412, 1424, 1438, 1454, 1472, 1492, 1514, 1528, 1543, 1551, 1557, 1571, 1586, 1596, 1612, 1627, 1637, 1645, 1652, 1661, 1674, 1690, 1705, 1714, 1725, 1734, 1743, 1756, 1765, 1778, 1791, 1798, 1808, 1823}

func (i actionType) String() string {
	if i < 0 || i >= actionType(len(_actionType_index)-1) {
		return "actionType(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _actionType_name[_actionType_index[i]:_actionType_index[i+1]]
}
